﻿using Google.Apis.Auth.OAuth2;
using Google.Apis.Services;
using Google.Apis.Upload;
using Google.Apis.YouTube.v3;
using Google.Apis.YouTube.v3.Data;
using Leadtools.Multimedia;
using LMDVRSinkLib;
using LMScrCapLib;
using LMVCallbackLib;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;

namespace LEADTOOLS_Game_Capture_Demo
{
   public partial class MainForm : Form
   {
      private ILMScrCap _scrCap;
      private LMDVRSink _dvrSink;
      private ScrCaptureCallback _callBack;
      private StillImageCallback _stillCallback;
      private LMVCallback _lmvCallback;
      private CaptureCtrl _captureCtrl;
      private System.Timers.Timer _timer;
      private string _outputDir, _videoOutputDir;
      private string _dvrFileName = "output.lbl";
      private string _googleApiJsonPath = "",
                     _youtubeVideoTitle = "Default Title",
                     _youtubeVideoDescription = "Default Description",
                     _youtubeChannelId = "",
                     _youtubeVideoPrivacyStatus = "Unlisted",
                     _youtubeVideoCategoryId = "20";
      private bool _stereoMixExist, _stillImaging;
      private int _videoClipLength = 30, _captureFPS = 30;
      private ToolTip Tooltip;
      private ResultDialog resultsDialog = new ResultDialog();
      public MainForm()
      {
         InitializeComponent();
         PrepareTooltips();

         _outputDir = Directory.GetCurrentDirectory();
         _videoOutputDir = Directory.GetCurrentDirectory();
         _captureCtrl = new CaptureCtrl(true);

         // Determine if we can capture audio from the stereo device
         _stereoMixExist = SetupAudioDevices();

         // If the Stereo Mix audio device is not present on the machine then screen capture is not possible 
         // so stop the application
         if (_stereoMixExist == false)
            MessageBox.Show("The Stereo Mix Audio Device is Not Present or is Not Enabled on This Machine.\nThe Application Will Not Capture Sound.\n\nCertain audio cards support capturing the sound from the sound card. \n\nIn your task bar right click on the volume manager. Click on recording devices. If you see a device named something like 'Stereo Mix' or 'Wave mix' this should be the output of your sound card. You may need to enable it. If it’s there and enabled, the the demo should be able to capture the audio from it.\n\nAfter enabling the audio device, you can restart the demo so the demo can capture from the sound card.", "Audio Device Missing or Not Enabled");

         // Set up the screen capture, capture control, and DVR
         if (!SetupVideoDevice() || !SetupCaptureCtrl() || !SetupDvrSink())
            Application.Exit();

         // Determine if setting up the callback filter was successful
         _stillImaging = InsertCallbackFilter();

         _timer = new System.Timers.Timer(1000); // Fires event every second to update the form
         _timer.Elapsed += timer_Elapsed;

         btnPlay.Enabled = false;
         btnPause.Enabled = false;
         btnStop.Enabled = false;
         btnFullVideo.Enabled = false;
         btnVideoClip.Enabled = false;
         btnScreenshot.Enabled = false;
      }

      private void btnPlay_Click(object sender, EventArgs e)
      {
         try
         {
            // Start the capture control based on if audio and video are enabled
            if (_stereoMixExist == true)
            {
               if (_captureCtrl.State == CaptureState.Paused)
                  _captureCtrl.RunCapture();
               else
                  _captureCtrl.StartCapture(CaptureMode.VideoAndAudio);
            }
            else
            {
               if (_captureCtrl.State == CaptureState.Paused)
                  _captureCtrl.RunCapture();
               else
                  _captureCtrl.StartCapture(CaptureMode.Video);
            }

            // Setup the timer for displaying the recording time period on the form
            _timer = new System.Timers.Timer(1000); // 10 seconds
            _timer.Elapsed += timer_Elapsed;
            _timer.Enabled = true;

            if (_stillImaging == true)
               btnScreenshot.Enabled = true;
            btnPause.Enabled = true;
            btnStop.Enabled = true;
            btnFullVideo.Enabled = true;
            btnVideoClip.Enabled = true;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      public delegate void UpdateLabelDelegate(string time);

      private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
      {
         try
         {
            TimeSpan t = TimeSpan.FromSeconds(_captureCtrl.CaptureTime);

            string answer = string.Format("{0:D2}:{1:D2}:{2:D2}", t.Hours, t.Minutes, t.Seconds);

            lblTimer.Invoke(new UpdateLabelDelegate(this.UpdateLabel), answer);
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void UpdateLabel(string time)
      {
         lblTimer.Text = time;
      }

      private void btnPause_Click(object sender, EventArgs e)
      {
         try
         {
            // Pause the capture control
            _captureCtrl.PauseCapture();
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void btnStop_Click(object sender, EventArgs e)
      {
         try
         {
            // Stop the capture control
            _captureCtrl.StopCapture();
            Reset();
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void btnFullVideo_Click(object sender, EventArgs e)
      {
         try
         {
            // Use the DVR buffer on disk to save the entire recorded video to an AVI file
            double first, last, total;
            // Get the current byte array information for the stream
            _dvrSink.GetAvailabilityInfo(out first, out last, out total);
            // Copy the data to the new file 
            _dvrSink.CopyBufferToFile(_videoOutputDir + @"\" + _callBack.WindowName + "_Full.avi", first, last);

            resultsDialog.ShowDialog(this, "Full Video Saved to:\n" + _videoOutputDir + @"\" + _callBack.WindowName + "_Full.avi", _videoOutputDir, "Full Video Saved");
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void btnVideoClip_Click(object sender, EventArgs e)
      {
         try
         {
            // User the DVR buffer on disk and the video clip length setting to save a video clip to disk
            double first, last, total, ratio, thirtySecondOffset, firstOffset;
            if (_captureCtrl.CaptureTime < _videoClipLength)
            {
               // Get the current byte array information for the stream
               _dvrSink.GetAvailabilityInfo(out first, out last, out total);
               // Copy the data to the new file 
               _dvrSink.CopyBufferToFile(_videoOutputDir + @"\" + _callBack.WindowName + "_Clip.avi", first, last);
               resultsDialog.ShowDialog(this, "Video Clip Saved to:\n" + _videoOutputDir + @"\" + _callBack.WindowName + "_Clip.avi", _videoOutputDir, "Video Clip Saved");

               return;
            }

            // Get the current byte array information for the stream
            _dvrSink.GetAvailabilityInfo(out first, out last, out total);
            // Calculate the ratio for video clip of the stream
            ratio = _videoClipLength / _captureCtrl.CaptureTime;
            // Calculate what the total offset should be to get a video clip
            thirtySecondOffset = total * ratio;
            // Calculate the first offset value based on the total offset for a video clip
            firstOffset = total - thirtySecondOffset;
            // Copy the data to the new file 
            _dvrSink.CopyBufferToFile(_videoOutputDir + @"\" + _callBack.WindowName + "_Clip.avi", firstOffset, last);

            resultsDialog.ShowDialog(this, "Video Clip Saved to:\n" + _videoOutputDir + @"\" + _callBack.WindowName + "_Clip.avi", _videoOutputDir, "Video Clip Saved");
         }
         catch (Exception ex)
         {
            MessageBox.Show(this, ex.Message);
         }
      }

      private void btnScreenShot_Click(object sender, EventArgs e)
      {
         try
         {
            // If still imaging is available, take a screen shot
            if (_stillImaging)
            {
               _stillCallback.StillImageName = _callBack.WindowName;
               _stillCallback.TakeScreenShot = true;
            }
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void btnSelectWindow_Click(object sender, EventArgs e)
      {
         try
         {
            _captureCtrl.StopCapture();
            Reset();

            // User the Screen Capture filter's SelectWindowTool method to select the window to record
            _callBack = new ScrCaptureCallback();
            _scrCap = _captureCtrl.GetSubObject(CaptureObject.VideoCaptureFilter) as ILMScrCap;
            _scrCap.CaptureFrameRate = _captureFPS;
            _scrCap.ChangeSettings(true);
            _scrCap.EnableOutline = true;
            _scrCap.EnableRegion = true;
            _scrCap.SetCallbackObject(_callBack, 0);
            _scrCap.SelectWindowTool();
            _scrCap.RegionRight = _callBack.ClassRegionRight;
            _scrCap.RegionBottom = _callBack.ClassRegionBottom;
            _scrCap.RegionLeft = _callBack.ClassRegionLeft;
            _scrCap.RegionTop = _callBack.ClassRegionTop;
            _scrCap.ChangeSettings(false);

            btnPlay.Enabled = true;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void Reset()
      {
         btnPlay.Enabled = false;
         btnPause.Enabled = false;
         btnStop.Enabled = false;
         btnFullVideo.Enabled = false;
         btnVideoClip.Enabled = false;
         btnScreenshot.Enabled = false;

         _timer.AutoReset = true;
         _timer.Stop();
         lblTimer.Text = string.Format("{0:D2}:{1:D2}:{2:D2}", 0, 0, 0);
         lblTimer.Refresh();
      }

      private bool SetupVideoDevice()
      {
         try
         {
            // Set the LEAD Screen Capture (2.0) filter as the capture device
            bool screenCaptureFilterPresent = false;
            for (int i = 0; i < _captureCtrl.VideoDevices.Count; i++)
            {
               if (_captureCtrl.VideoDevices[i].ToString() == "LEAD Screen Capture (2.0)")
               {
                  screenCaptureFilterPresent = true;
                  _captureCtrl.VideoDevices[i].Selected = true;
               }
            }
            // If the LEAD Screen Capture (2.0) filter is not present on the machine then screen capture is not possible 
            // so stop the application
            if (screenCaptureFilterPresent == false)
            {
               MessageBox.Show("The LEAD Screen Capture (2.0) Filter is Not Present on This Machine.\nThe Application Will Now Close", "LEAD Screen Capture (2.0) Filter Not Found");
               return false;
            }

            return true;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
            return false;
         }
      }

      private bool SetupAudioDevices()
      {
         try
         {
            // Set the Stereo Mix as the audio device
            for (int i = 0; i < _captureCtrl.AudioDevices.Count; i++)
            {
               Console.WriteLine(_captureCtrl.AudioDevices[i].FriendlyName);
               if (_captureCtrl.AudioDevices[i].FriendlyName.Contains("Mix"))
               {
                  _captureCtrl.AudioDevices[i].Selected = true;
                  return true;
               }
            }
            return false;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
            return false;
         }
      }

      private bool SetupCaptureCtrl()
      {
         try
         {
            // Set the output format
            _captureCtrl.VideoCompressors.H264.Selected = true;
            if (_stereoMixExist == true)
               _captureCtrl.AudioCompressors.AAC.Selected = true;
            _captureCtrl.TargetFormat = TargetFormatType.DVRTransport;
            _captureCtrl.TargetFile = _dvrFileName;

            if (_stereoMixExist == true)
               _captureCtrl.ReadyCapture(CaptureMode.VideoAndAudio);
            else
               _captureCtrl.ReadyCapture(CaptureMode.Video);

            _captureCtrl.TargetFormats[TargetFormatType.DVRTransport].UseFilterCache = true;

            return true;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
            return false;
         }
      }

      private void PrepareTooltips()
      {
         Tooltip = new ToolTip();
         Tooltip.InitialDelay = 100;
         Tooltip.AutomaticDelay = 100;
         Tooltip.ShowAlways = true;
         foreach (Control ctrl in Controls)
         {
            if (ctrl is Button && ctrl.Tag is string)
            {
               ctrl.MouseHover += new EventHandler(delegate (object o, EventArgs a)
               {
                  var btn = (Control)o;
                  Tooltip.SetToolTip(btn, btn.Tag.ToString());
               });

               ctrl.EnabledChanged += new EventHandler(delegate (object o, EventArgs a)
               {
                  switch (ctrl.Name)
                  {
                     case "btnFullVideo":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.full_video_disabled : Properties.Resources.btnFullVideo;
                        break;
                     case "btnPause":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.pause_disabled : Properties.Resources.btnPause;
                        break;
                     case "btnPlay":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.play_disabled : Properties.Resources.btnPlay;
                        break;
                     case "btnScreenshot":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.screenshot_disabled : Properties.Resources.btnScreenshot;
                        break;
                     case "btnStop":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.stop_disabled : Properties.Resources.btnStop;
                        break;
                     case "btnVideoClip":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.video_clip_disabled : Properties.Resources.btnVideoClip;
                        break;
                     case "btnSelectWindow":
                        ctrl.BackgroundImage = ctrl.Enabled == false ? Properties.Resources.select_window_disabled : Properties.Resources.btnSelectWindow;
                        break;
                     default:
                        break;
                  }
               });
            }
         }
      }

      private async void uploadToolStripMenuItem_Click(object sender, EventArgs e)
      {
         try
         {
            // Make sure all the necessary Youtube information has been provided for video upload to function correctly
            if (_googleApiJsonPath == "" || _youtubeVideoTitle == "" || _youtubeVideoDescription == "" || _youtubeChannelId == "" || _youtubeVideoPrivacyStatus == "" || _youtubeVideoCategoryId == "")
            {
               MessageBox.Show(this, "All Youtube Information is not filled out.\nPlease Enter your Youtube Information.\n\nFile -> Upload to Youtube -> Settings", "Youtube Information Missing");
               return;
            }

            var filePath = _videoOutputDir + @"\" + _callBack.WindowName + "_Clip.avi";
            if (!File.Exists(filePath))
            {
               MessageBox.Show("Video clip does not exist to upload.\nPlease save a video clip before attempting to upload");
               return;
            }

            UserCredential credential;
            using (var stream = new FileStream(_googleApiJsonPath, FileMode.Open, FileAccess.Read))
            {
               credential = await GoogleWebAuthorizationBroker.AuthorizeAsync(
                   GoogleClientSecrets.Load(stream).Secrets,
                   // This OAuth 2.0 access scope allows an application to upload files to the
                   // authenticated user's YouTube channel, but doesn't allow other types of access.
                   new[] { YouTubeService.Scope.YoutubeUpload },
                   "user",
                   CancellationToken.None
               );
            }

            // Initialize the YouTubeService
            var youtubeService = new YouTubeService(new BaseClientService.Initializer()
            {
               HttpClientInitializer = credential,
               ApplicationName = Assembly.GetExecutingAssembly().GetName().Name
            });

            // Create the video object based on the video clip taken
            var video = new Video();
            video.Snippet = new VideoSnippet();
            video.Snippet.Title = _youtubeVideoTitle;
            video.Snippet.Description = _youtubeVideoDescription;
            video.Snippet.CategoryId = _youtubeVideoCategoryId;
            video.Snippet.ChannelId = _youtubeChannelId;
            video.Status = new VideoStatus();
            video.Status.PrivacyStatus = _youtubeVideoPrivacyStatus;

            // Upload the video to youtube using the user's settings
            using (var fileStream = new FileStream(filePath, FileMode.Open))
            {
               var videosInsertRequest = youtubeService.Videos.Insert(video, "snippet,status", fileStream, "video/*");
               videosInsertRequest.ProgressChanged += videosInsertRequest_ProgressChanged;
               videosInsertRequest.ResponseReceived += videosInsertRequest_ResponseReceived;

               await videosInsertRequest.UploadAsync();
            }
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void videosInsertRequest_ProgressChanged(IUploadProgress progress)
      {
         try
         {
            switch (progress.Status)
            {
               case UploadStatus.Uploading:
                  Console.WriteLine("{0} bytes sent.", progress.BytesSent);
                  break;

               case UploadStatus.Failed:
                  Console.WriteLine("An error prevented the upload from completing.\n{0}", progress.Exception);
                  break;
            }
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private void videosInsertRequest_ResponseReceived(Video video)
      {
         MessageBox.Show("Video '" + _youtubeVideoTitle + "' was successfully uploaded.");
      }

      private void exitToolStripMenuItem_Click(object sender, EventArgs e)
      {
         this.Close();
      }

      private void captureSettingsToolStripMenuItem_Click(object sender, EventArgs e)
      {
         // Show the VideoSettingsDialog to get the information from the user
         VideoSettingsDialog dlg = new VideoSettingsDialog(_captureFPS, _videoClipLength, _videoOutputDir);
         dlg.ShowDialog();

         _captureFPS = dlg.CaptureFps;
         _videoClipLength = dlg.VideoClipLength;
         _videoOutputDir = dlg.VideoSaveDirectory;
      }

      private void youtubeUploadSettingsToolStripMenuItem1_Click(object sender, EventArgs e)
      {
         try
         {
            // Show the UploadYoutubeSettings dialog to get the user's Youtube information
            UploadYoutubeSettings youtubeSettings = new UploadYoutubeSettings(_googleApiJsonPath, _youtubeVideoTitle, _youtubeVideoDescription, _youtubeChannelId, _youtubeVideoPrivacyStatus, _youtubeVideoCategoryId);
            youtubeSettings.ShowDialog();

            _googleApiJsonPath = youtubeSettings.JsonFilePath;
            _youtubeVideoTitle = youtubeSettings.VideoTitle;
            _youtubeVideoDescription = youtubeSettings.VideoDescription;
            _youtubeChannelId = youtubeSettings.ChannelId;
            _youtubeVideoPrivacyStatus = youtubeSettings.VideoPrivacyStatus;
            _youtubeVideoCategoryId = youtubeSettings.VideoCategoryId;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
         }
      }

      private bool SetupDvrSink()
      {
         try
         {
            // Get the DVR Sink object from the control
            _dvrSink = (LMDVRSink)_captureCtrl.TargetFormats[TargetFormatType.DVRTransport].GetCacheObject(TargetFormatObject.Sink);
            // Tell the DVR Sink that settings are about to change
            _dvrSink.StartChangingAttributes();
            // Set only one DVR buffer folder 
            _dvrSink.FolderCount = 1;
            // Set the buffer base file name to 'Capture.LBL' 
            _dvrSink.BaseName = "output.lbl";
            // Set the buffer folder location to 'C:\Temp\DVR' 
            _dvrSink.set_FolderName(0, _outputDir);
            // Set the buffer folder to have 5 buffer data files, each at 100MB max size 
            _dvrSink.SetBufferSize(0, 10, 102400000);
            // Commit the changed settings now 
            _dvrSink.StopChangingAttributes(false);

            return true;
         }
         catch (Exception ex)
         {
            MessageBox.Show(ex.Message);
            return false;
         }
      }

      private bool InsertCallbackFilter()
      {
         try
         {
            // Set up the callback filter
            Processor videoCallback = _captureCtrl.VideoProcessors.Callback;
            _captureCtrl.SelectedVideoProcessors.Add(videoCallback);
            _lmvCallback = (LMVCallback)_captureCtrl.GetSubObject(CaptureObject.SelVideoProcessor);

            _stillCallback = new StillImageCallback();
            _lmvCallback.ReceiveProcObj = _stillCallback;

            return true;
         }
         catch (Exception ex)
         {
            MessageBox.Show("An error has occured in adding the filter\n\n" + ex.Message + "\n\nStill image capturing is disabled.", "Error Inserting Callback Filter");
            return false;
         }
      }
   }
}
